package server

import (
	"bytes"
	"context"
	"encoding/json"
	"fmt"
	"log"
	"os"
	"strconv"
	"time"

	"github.com/go-redis/redis/v8"
	"github.com/jackc/pgx/v5/pgxpool"
	"github.com/streadway/amqp"

	"git.pbiernat.dev/egommerce/go-api-pkg/consul"
	"git.pbiernat.dev/egommerce/go-api-pkg/fluentd"

	db "git.pbiernat.dev/egommerce/basket-service/pkg/database"
	srv "git.pbiernat.dev/egommerce/basket-service/pkg/server"

	"git.pbiernat.dev/egommerce/basket-service/internal/common"
)

type (
	Server struct {
		Base   *srv.Server
		Config *common.Config

		Cache    *redis.Client
		Database *pgxpool.Pool
		Eventbus *amqp.Channel
		Logger   *fluentd.Logger
		Registry *consul.Service

		kvNmspc string
	}

	OptionFn func(*Server) error
)

func New(c *common.Config, opts ...OptionFn) *Server {
	svr := &Server{
		Base: srv.New(c.Base),
	}

	for _, opt := range opts {
		if err := opt(svr); err != nil {
			log.Fatalf("Failed to start HTTP Server. Err: %v\n", err)
		}
	}

	SetupMiddleware(svr)
	SetupRouter(svr)

	return svr
}

func WithCache(c *common.Config) OptionFn {
	redis := redis.NewClient(&redis.Options{
		Addr:     c.CacheAddr,
		Password: c.CachePassword,
		DB:       0,
	})

	return func(s *Server) error {
		s.Cache = redis

		return nil
	}
	// defer redis.Close()
}

func WithDatabase(c *common.Config) OptionFn {
	dbConn, err := db.Connect(c.DbURL)
	if err != nil {
		log.Fatalf("Failed to connect to the Database: %s. Err: %v\n", c.DbURL, err)
		os.Exit(1)
	}

	return func(s *Server) error {
		s.Database = dbConn

		return nil
	}
	// defer dbConn.Close()
}

func WithEventbus(c *common.Config) OptionFn {
	conn, err := amqp.Dial(c.EventBusURL)
	if err != nil {
		log.Fatalf("Failed to connect to the Eventbus: %s. Err: %v\n", c.EventBusURL, err)
	}

	chn, err := conn.Channel()
	if err != nil {
		log.Fatalf("Failed to open new Eventbus channel. Err: %v\n", err)
	}

	return func(s *Server) error {
		s.Eventbus = chn

		return nil
	}

	// defer ebCh.Close()
	// defer amqp.Close(ebConn)
}

func WithLogger(c *common.Config) OptionFn {
	logHost, logPort, err := fluentd.ParseAddr(c.LoggerAddr)
	if err != nil {
		log.Fatalf("Failed to parse Fluentd address: %s. Err: %v", c.LoggerAddr, err)
	}

	logger, err := fluentd.NewLogger(c.GetAppFullName(), logHost, logPort)
	if err != nil {
		log.Fatalf("Failed to connect to the Fluentd on %s:%d. Err: %v", logHost, logPort, err)
	}

	return func(s *Server) error {
		s.Logger = logger

		return nil
	}
	// defer logger.Close()
}

func WithRegistry(c *common.Config) OptionFn {
	// fmt.Printf("WithRegistry constructor: config: %v", c.Base)
	port, _ := strconv.Atoi(c.Base.NetAddr[1:]) // FIXME: can be IP:PORT which now will cause error
	registry, err := consul.NewService(c.RegistryAddr, c.Base.AppID, c.Base.AppName, c.Base.AppID, c.Base.AppName, c.Base.PathPrefix, port)
	if err != nil {
		log.Fatalf("Failed to connect to the Consul on: %s. Err: %v", c.RegistryAddr, err)
	}

	err = registry.Register()
	if err != nil {
		log.Fatalf("Failed to register in the Consul service. Err: %v", err)
	}

	return func(s *Server) error {
		s.Registry = registry

		go func(s *Server) { // Consul KV updater
			ticker := time.NewTicker(time.Second * 15)
			for range ticker.C {
				s.updateKVConfig()
			}
		}(s)

		go func(s *Server) { // Server metadata cache updater
			ticker := time.NewTicker(time.Second * 5)
			for range ticker.C {
				s.cacheMetadata()
			}
		}(s)

		return nil
	}
}

// REFACTOR IN PROGRESS
// func (s *Server) Shutdown() error {
// 	s.Logger.Log("Server is going down... Unregistering service: %s", s.Discovery.GetID())
// 	s.Discovery.Unregister()
// 	s.clearMetadataCache()

// 	s.Cache.Close()
// 	s.Database.Close()
// 	s.Eventbus.Close()
// 	s.Logger.Close()

// 	return nil
// }

func (s *Server) Shutdown() srv.PurgeFn {
	return func(srv *srv.Server) error {
		fmt.Printf("%v", s.Base)
		// s.Logger.Log("Server is going down... Unregistering service: %s", s.Base.AppID)
		s.Logger.Log("Server is going down... Unregistering service...")

		s.Registry.Unregister()
		s.clearMetadataCache()
		s.Eventbus.Close()
		s.Database.Close()
		s.Logger.Close()

		return s.Base.Shutdown()
	}
}

// END: REFACTOR IN PROGRESS

// @CHECK: merge s.Config and s.Base.Config to display all config as one array/map
func (s *Server) updateKVConfig() {
	config, _, err := s.Registry.KV().Get(s.kvNmspc, nil)
	if err != nil || config == nil {
		return
	}

	kvCnf := bytes.NewBuffer(config.Value)
	decoder := json.NewDecoder(kvCnf)
	if err := decoder.Decode(&s.Config); err != nil {
		return
	}
}

func (s *Server) cacheMetadata() {
	ctx := context.TODO()
	key, address := s.getMetadataIPsKey(), s.Base.Config.AppID

	pos := s.Cache.LPos(ctx, key, address, redis.LPosArgs{}).Val()
	if pos >= 0 {
		s.Cache.LRem(ctx, key, 0, address)
	}

	s.Cache.LPush(ctx, key, address).Err()
}

func (s *Server) clearMetadataCache() {
	ctx := context.TODO()
	fmt.Printf("metadata: %v", s.Config.Base)
	key, address := s.getMetadataIPsKey(), s.Config.Base.AppID

	s.Cache.LRem(ctx, key, 0, address)
}

func (s *Server) getMetadataIPsKey() string {
	return "internal__" + s.Base.Config.AppName + "__ips"
}

//
//
//
//// OLD CODE TO BE REMOVED
// func NewServer(conf *Config, logger *fluentd.Logger, db *pgxpool.Pool, cache *redis.Client, ebCh *amqp.Channel) *Server {

// cnf := fiber.Config{
// 	AppName:      conf.AppName,
// 	ServerHeader: conf.AppName,
// 	ReadTimeout:  conf.ReadTimeout * time.Millisecond,
// 	WriteTimeout: conf.WriteTimeout * time.Millisecond,
// 	IdleTimeout:  conf.IdleTimeout * time.Millisecond,
// }
// s := &Server{
// 	fiber.New(cnf),
// 	conf,
// 	logger,
// 	db,
// 	cache,
// 	ebCh,
// 	consul,
// 	conf.AppName,
// 	conf.NetAddr,
// 	conf.KVNamespace,
// }

// return s
// }
